// ppcbf.cxx - Implementations of hand-written functions for the PowerPC
// simulator. -*- C++ -*-

// Copyright (C) 2002 Red Hat, Inc.
// Written by Johan Rydberg, jrydberg@rtmk.org

// This file is part of SID and is licensed under the GPL.
// See the file COPYING.SID for conditions for redistribution.

#include "ppcbf.h"

using namespace std;
using namespace sid;
using namespace sidutil;
using namespace ppcbf;

ppcbf_cpu::ppcbf_cpu (): engine (32768) // XXX: tune size
{
  for (unsigned i = 0; i < 32; ++i)
    {
      string name = string ("r") + make_numeric_attribute (i);
      SI* value = & this->hardware.h_gpr[i];
      this->add_watchable_register (name, value);
    }

  {
    string name = "pc";
    USI* value = & this->hardware.h_pc;
    this->add_watchable_register (name, value);
  }

  this->create_gdb_register_attrs (70, "", & this->hardware.h_pc);
}


void
ppcbf_cpu::reset ()
{
  int i;

  // ??? this is not correct.
  this->hardware.h_pc = 0x100;
  for (i = 0; i < 32; i++)
    this->hardware.h_gpr[i] = 0;
}


void
ppcbf_cpu::flush_icache ()
{
  this->engine.flush ();
}


void
ppcbf_cpu::invalid_insn (PCADDR pc)
{
  // Update recorded pc in case software trap handling component wants to
  // use it.  This is also needed if the insn is handled as hardware would.
  this->h_pc_set (pc);

  cpu_trap_disposition whatnext 
    = this->signal_trap (cpu_trap_invalid_insn, pc);

  switch (whatnext)
    {
    case cpu_trap_unhandled:
      cerr << "hw-cpu-ppc: invalid insn not implemented" << endl;
      break;

    case cpu_trap_skip:
      {
	PCADDR npc = pc + 4;
	this->h_pc_set (npc);
      }
      break;

    case cpu_trap_handled:
    case cpu_trap_reissue:
      break;

    default:
      abort ();
    }

  throw cpu_exception ();
}


// Memory trap(/fault) handling.
void
ppcbf_cpu::memory_trap (const cpu_memory_fault& t)
{
  this->h_pc_set (t.pc);

  cpu_trap_disposition whatnext 
    = this->signal_trap (cpu_trap_memory_fault, t.address);

  switch (whatnext)
    {
    case cpu_trap_unhandled:
      cerr << "hw-cpu-ppc: memory fault handler unimplemented!" << endl;
      break;

    case cpu_trap_skip:
      {
	PCADDR pc = t.pc + 4;
	this->h_pc_set (pc);
      }
      break;

    case cpu_trap_handled:
    case cpu_trap_reissue:
      break;

    default:
      abort ();
    }
}

SI
ppcbf_cpu::ppcbf_trap (PCADDR pc, USI trap)
{
  // Update recorded pc in case software trap handling component wants to
  // use it.  This is also needed if the insn is handled as hardware would.
  this->h_pc_set (pc);

  cpu_trap_disposition whatnext 
    = this->signal_trap (cpu_trap_software, trap);

  switch (whatnext)
    {
    case cpu_trap_unhandled:
      cerr << "hw-cpu-ppc: trap insn not implemented" << endl;
      break;

    case cpu_trap_skip:
      {
	PCADDR npc = pc + 4;
	this->h_pc_set (npc);
      }
      break;

    case cpu_trap_handled:
    case cpu_trap_reissue:
      break;

    default:
      abort ();
    }

  throw cpu_exception ();
}

SI
ppcbf_cpu::ppcbf_rfi (void)
{
  // ??? set msr
  return (h_spr_get (27) & ~3);
}


// Stepper

void
ppcbf_cpu::step_insns ()
{
  while (true)
    {
      // Fetch/decode the instruction  ------------------------------
      PCADDR pc = this->h_pc_get ();
      bool found;
      ppc_scache* sem = this->engine.find (pc, found);

      if (! found)
	{
	  USI insn;

	  try
	    {
	      insn = this->GETIMEMSI (pc, pc);
	    }
	  catch (cpu_memory_fault& t)
	    {
	      this->memory_trap (t);
	      break;
	    }
	      
	  sem->decode (this, pc, insn, insn);
	}

      // ??? check so that we do not try to execute a supervisor insn
      //     in user mode. 

      // Execute the instruction  -----------------------------------
      if (this->trace_result_p)
	this->begin_trace (pc, sem->idesc->insn_name);
      try
	{
	  sem->idesc->execute (this, sem);
	}
      catch (cpu_memory_fault& t)
	{
	  this->memory_trap (t);
	  this->yield ();
	}
      catch (cpu_exception& t)
	{
	  this->yield ();
	}
      if (this->trace_result_p)
	this->end_trace ();
      
      // Do post-instruction processing  ----------------------------
      if (this->enable_step_trap_p) 
	this->signal_trap (sidutil::cpu_trap_stepped);
      this->triggerpoint_manager.check_and_dispatch ();
      
      // test for exit condition
      if (stop_after_insns_p (1))
	break;
    }
}

// State save/restore

void 
ppcbf_cpu::stream_state (ostream& o) const
{
  // call base class first
  cgen_bi_endian_cpu::stream_state (o);

  o << " ppcbf-cpu";

  // XXX: Cheat by dumping the entire hardware struct in binary form.  That
  // means that this state image will not be restorable on a different
  // endianness host platform.
  string hw = string (reinterpret_cast<const char*> (& this->hardware), 
		      sizeof (this->hardware));
  o << " " << string2stream (hw);
}

void 
ppcbf_cpu::destream_state (istream& i)
{
  // call base class first
  cgen_bi_endian_cpu::destream_state (i);

  string key;
  i >> key;
  if (key != "ppcbf-cpu")
    {
      i.setstate (ios::badbit);
      return;
    }

  string hw;
  i >> stream2string(hw);
  if (hw.size() != sizeof(this->hardware))
    i.setstate (ios::badbit);
  else
    {
      const char* hwbin = hw.data();
      ::memcpy (reinterpret_cast<char*>(& this->hardware), 
		hwbin, sizeof(this->hardware));
    }
}


// Debugger interface functions: return bitwise register images in
// "target byte order".

string
ppcbf_cpu::dbg_get_reg (host_int_4 reg)
{
  string attr;
 
  // 32-bit registers;
  if (reg < 32 || reg > 63)
    {
      host_int_4 value;

      // FIXME this register number mapping must match gdb's.
      if (reg < 32) // general purpose registers
	value = h_gpr_get (reg);
      else if (reg == 64)
	value = h_pc_get ();
      else if (reg == 66)
	value = h_cr_get ();
      else if (reg == 67)
	value = h_lr_get ();
      else if (reg == 68)
	value = h_ctr_get ();
      else if (reg == 69)
	value = 0; /* xer */
      else
	value = 0xdeadbeef;
      
      // change to "target endian"
      big_int_4 v = value;
      for (unsigned i=0; i<4; i++)
	attr += v.read_byte (i);
    }
  else // 64-bit regs (floating point)
    { 
      host_int_8 value = 0xdeadbeef;

      // change to "target endian"
      big_int_8 v = value;
      for (unsigned i=0; i<8; i++)
	attr += v.read_byte (i);
    }
  return attr;
}

component::status 
ppcbf_cpu::dbg_set_reg (host_int_4 reg, const string& attr)
{
  // change from "target endian"
  big_int_4 v;
  for (unsigned i=0; i < 4; i++)
    v.write_byte (i, attr[i]);
  host_int_4 value = v;

  // This register number mapping must match gdb's.
  if (reg < 32) // general purpose registers
    h_gpr_set (reg, value);
  else if (reg == 64)
    h_pc_set (value);
  else
    ;

  return component::ok;
}
