// ppc.cxx - Implementation of the PPC libgloss/ppcbug specifics.  -*- C++ -*-

// Copyright (C) 2002 Red Hat, Inc.
// Written by Johan Rydberg, jrydberg@rtmk.org

// This file is part of SID and is licensed under the GPL.
// See the file COPYING.SID for conditions for redistribution.

#include "config.h"
#include "ppc.h"
#include "libgloss.h"
#include "newlib.h"

using namespace std;
using namespace sid;
using namespace sidutil;

ppc_libgloss::ppc_libgloss()
{
  register_attribute_names[-1] = "r0"; // result
  register_attribute_names[-2] = "r0"; // error
  register_attribute_names[0] = "r0";  // function code
  register_attribute_names[1] = "r3";  // arg 1
  register_attribute_names[2] = "r4";  // arg 2
  register_attribute_names[3] = "r5";  // arg 3
}

bool
ppc_libgloss::get_int_argument(unsigned index, int32& value)
{
  assert (cpu);
  string attrName = register_attribute_names[index];
  assert (attrName != "");

  string attrValue = cpu->attribute_value (attrName);
  if (attrValue == "")
    {
      cerr << "Could not read attribute " << attrName
	   << " for ABI argument #" << index << endl;
      return false;
    }

  host_int_4 value_number;
  parse_attribute (attrValue, value_number);
  value = value_number;
  return true;
}

bool
ppc_libgloss::set_int_result(int32 value) 
{
  assert (cpu);
  string attrName = register_attribute_names[-1];
  assert (attrName != "");

  host_int_4 value_number = value;
  string attrValue = make_attribute (value_number);

  cpu->set_attribute_value (attrName, attrValue);
  return true; // XXX: check?
}

bool
ppc_libgloss::set_error_result(int32 value) 
{
  assert (cpu);
  string attrName = register_attribute_names[-2];
  assert (attrName != "");

  host_int_4 value_number = value;
  string attrValue = make_attribute (value_number);

  cpu->set_attribute_value (attrName, attrValue);
  return true; // XXX: safe to assume success?
}

bool
ppc_libgloss::syscall_trap_p()
{
  // PPC system calls use TRAP 0.
  return ((trap_type_ipin.sense() == sidutil::cpu_trap_software)
	  && (trap_code_pin.sense() == 0xc)); /* trap 0xc */
}

void
ppc_libgloss::syscall_trap()
{
  int32 syscall;

  get_int_argument(0, syscall);
  if (verbose_p)
    {
      cerr << "ppc/libgloss system call number " << syscall << endl;
    }

  switch (syscall)
    {
    case libgloss::SYS_exit:
      {
	int32 value;
	get_int_argument (1, value);
	if (verbose_p)
	  cerr << "*** exit(" << value << ")" << endl;

	if (value == 0)
	  process_signal_pin.drive (newlib::sigQuit);
	else
	  process_signal_pin.drive (newlib::sigAbrt);
      }
      break;


    case 4:
      {
	int32 fd;
	int32 address;
	int32 length;
	string str;

	get_int_argument (1, fd);
	get_int_argument (2, address);
	get_int_argument (3, length);

	size32 len_written;
	int errcode;
	if (! this->write (fd, address, length, len_written, errcode))
	  {
	    set_error_result (host_to_target_errno(errcode));
	    set_int_result (-1);
	    return;
	  }

	set_int_result (len_written);
      }
      break;

    case 3:
      {
	int32 fd;
	int32 address;
	int32 length;
	string str;

	get_int_argument (1, fd);
	get_int_argument (2, address);
	get_int_argument (3, length);

	size32 len_read;
	int errcode;
	if (! this->read (fd, address, length, len_read, errcode))
	  {
	    set_error_result (host_to_target_errno(errcode));
	    set_int_result (-1);
	    return;
	  }

	set_int_result (len_read);
      }
      break;

    case 17:
      {
	int32 addr;

	get_int_argument (1, addr);

	if (verbose_p)
	  cerr << "ppc/libgloss brk " << hex << addr << dec << endl;
	
	set_int_result (0);
      }
      break;

    default:
      if (verbose_p)
	{
	  cerr << "Unimplemented syscall " << syscall << endl;
	}
      set_int_result (-1);
      set_error_result (newlib::eNoSys);
    }
}

// Convert host errno value to target.

int
ppc_libgloss::host_to_target_errno (int host_errno)
{
  return newlib::host_to_target_errno (host_errno);
}


ppc_ppcbug::ppc_ppcbug()
{
  register_attribute_names[-1] = "r3"; // result
  register_attribute_names[-2] = "r0"; // error
  register_attribute_names[0] = "r10";  // function code
  register_attribute_names[1] = "r3";   // arg 1
  register_attribute_names[2] = "r4";   // arg 2
  register_attribute_names[3] = "r5";   // arg 3
}

bool
ppc_ppcbug::get_int_argument(unsigned index, int32& value)
{
  assert (cpu);
  string attrName = register_attribute_names[index];
  assert (attrName != "");

  string attrValue = cpu->attribute_value (attrName);
  if (attrValue == "")
    {
      cerr << "Could not read attribute " << attrName
	   << " for ABI argument #" << index << endl;
      return false;
    }

  host_int_4 value_number;
  parse_attribute (attrValue, value_number);
  value = value_number;
  return true;
}

bool
ppc_ppcbug::set_int_result(int32 value) 
{
  assert (cpu);
  string attrName = register_attribute_names[-1];
  assert (attrName != "");

  host_int_4 value_number = value;
  string attrValue = make_attribute (value_number);

  cpu->set_attribute_value (attrName, attrValue);
  return true; // XXX: check?
}

bool
ppc_ppcbug::set_error_result(int32 value) 
{
  assert (cpu);
  string attrName = register_attribute_names[-2];
  assert (attrName != "");

  host_int_4 value_number = value;
  string attrValue = make_attribute (value_number);

  cpu->set_attribute_value (attrName, attrValue);
  return true; // XXX: safe to assume success?
}

bool
ppc_ppcbug::syscall_trap_p()
{
  // PPC system calls use TRAP 0.
  return ((trap_type_ipin.sense() == sidutil::cpu_trap_software)
	  && (trap_code_pin.sense() == 0xc)); /* trap 0xc */
}

void
ppc_ppcbug::syscall_trap()
{
  int32 syscall;

  get_int_argument(0, syscall);
  if (verbose_p)
    {
      cerr << "ppc/ppcbug system call number " << syscall << endl;
    }

  switch (syscall)
    {
    case ppc_ppcbug::BUGAPI_inchr:
      {
	unsigned char ch = -1;
	int length, errcode;

	set_int_result (ch);
      }
      break;

    case ppc_ppcbug::BUGAPI_inln:
      {
	int32 address;
	
	get_int_argument (1, address);

	size32 len_read;
	int errcode;
	if (! this->read (0, address, 256, len_read, errcode))
	  {
	    set_error_result (host_to_target_errno (errcode));
	    set_int_result (-1);
	    return;
	  }

	set_int_result (address + len_read);
      }
      break;

    case ppc_ppcbug::BUGAPI_time:
      set_int_result (time (0));
      break;

    case ppc_ppcbug::BUGAPI_outchr:
      {
	host_int_4 value;
	int32 ch;

	get_int_argument (1, ch);

	value = ch;
	tx_pin.drive (value);
      }
      break;

    case ppc_ppcbug::BUGAPI_outln:
      {
	int32 fd = 1;  // output device
	int32 address, stop;
	int32 length;
	string str;

	get_int_argument (1, address);
	get_int_argument (2, stop);

	length = (int) (stop - address);

	size32 len_written;
	int errcode;
	if (! this->write (fd, address, length, len_written, errcode))
	  {
	    set_error_result (host_to_target_errno(errcode));
	    set_int_result (-1);
	    return;
	  }
	else
	  {
	    host_int_4 value;

	    value = '\n';
	    tx_pin.drive (value);
	  }

	set_int_result (len_written);
      }
      break;
      
    case ppc_ppcbug::BUGAPI_outstr:
      {
	int32 fd = 1;  // output device
	int32 address, stop;
	int32 length;
	string str;

	get_int_argument (1, address);
	get_int_argument (2, stop);

	length = (int) (stop - address);

	size32 len_written;
	int errcode;
	if (! this->write (fd, address, length, len_written, errcode))
	  {
	    set_error_result (host_to_target_errno(errcode));
	    set_int_result (-1);
	    return;
	  }

      }
      break;

    case ppc_ppcbug::BUGAPI_return:
      {
	int32 value;
	get_int_argument (1, value);
	if (verbose_p)
	  cerr << "*** return(" << value << ")" << endl;

	if (value == 0)
	  process_signal_pin.drive (newlib::sigQuit);
	else
	  process_signal_pin.drive (newlib::sigAbrt);
      }
      break;

    default:
      if (verbose_p)
	{
	  cerr << "Unimplemented syscall " << syscall << endl;
	}
      set_int_result (-1);
      set_error_result (newlib::eNoSys);
    }
}

// Convert host errno value to target.

int
ppc_ppcbug::host_to_target_errno (int host_errno)
{
  return newlib::host_to_target_errno (host_errno);
}
